/*=========================================================================*
 *	File    :	th.h
 *
 *	Author  :	Pallav Gupta
 *	Created :	Sun Sep 14 20:44:25 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: th.h,v 1.15 2005/06/03 18:01:48 sis Exp $ */

#ifndef TH_H
#define TH_H

/* data structure of a threshold node */
typedef struct thnode {
  int nin;         /* number of inputs to this node */
  double *W;       /* array of input weights x0 -> w[0] etc. */
  double T;        /* threshold */
  double area;     /* area of this node */

  /* used by the thatpg command */
  int *fanin_lines;             /* fanin lines 0 = 1st fanout, 1 = 2nd fanout, so on */
  pset_family *pcubes;          /* array of propagation cubes for each input for this threshold node */
  pset_family *scovers;         /* singular covers, f[0] contains on cubes f[1] conains off cubes*/
  unsigned int simchanged:1;

  double scratch;        /* to be used as a scratch (temp) variable */
  
 /* add other data fields as necessary and modify thnode_check() if necessary*/
} thnode_t;


/* data structure that holds a splitted node in threshold logic synthesis */
typedef struct node_split_struct {
  int split_flag; /* flag to indicate the type of splitting: 1: AND, 0: OR */ 
  char *pn_name;  /* name of the parent node */
  array_t *nodes; /* array of splitted nodes */
} node_split_t;

/* options data structure for threshold synthesis */
typedef struct thsyn_options {
  bool help;      /* show help message */
  int max_fanin;  /* set the fanin restriction */
  int delta_on;   /* set the threshold on_offset */
  int delta_off;  /* set the threshold off_offset */
  char *filename;
} thsyn_options_t;

/* options data structure for threshold synthesis */
typedef struct majsyn_options {
  bool help;      /* show help message */
  int m;          /* set the collapse fanin restriction */
} majsyn_options_t;


/* options data structure for logic simulation */
typedef struct thsim_options {
  bool do_logic;     /* do combinational logic simulation */
  bool do_stg;       /* do STG simulation */
  bool help;         /* show help message */
  char *ifile;       /* read vectors from file */
  char *ofile;       /* dump simulation result to file */
  FILE *ip;          /* file pointer to ifile */
  FILE *op;          /* file pointer to ofile */
} thsim_options_t;


/* options data structure for thvcg command */
typedef struct thvcg_options {
  bool help;         /*show help message */
  int mode;          /* set the printing mode of vcg graph */
  int shrink;        /* shrink the node size */
  char *filename;    /* filename to write the vcg file */
} thvcg_options_t;

/* options data structure for thdot command */
typedef struct thdot_options {
  bool help;       /* show help message */
  int mode;        /* set the printing mode of dot graph */
  char *filename;  /* filename to write the dot file */
} thdot_options_t;

/* ALL THIS STUFF BELOW IS FOR THRESHOLD FAULT SIMULATION AND TEST GENERATION */
typedef enum thfault_value {
  TH_ATPG_SA0,                            /* perm SA0 */
  TH_ATPG_SA1                             /* perm SA1 */ 
} thfault_value_t;

typedef enum thfault_status {
  TH_ATPG_TESTED,                         /* tested fault */
  TH_ATPG_UNTESTED,                       /* initially all faults set to untested */
  TH_ATPG_ABORTED,                        /* couldn't find a test for this fault */
  TH_ATPG_REDUNDANT                       /* fault is untestable (i.e. redundant) */
} thfault_status_t;

typedef enum thfault_type {
  TH_ATPG_STEM,                           /* fault on a fanout node, stem fault */
  TH_ATPG_BRANCH,                         /* fault on one of the branches of a fanout node */
  TH_ATPG_NORMAL                          /* normal fault, i.e. no fanout */
} thfault_type_t;

/* 6V system for test generation */
typedef enum thatpg_simv_enum { 
  TH_ATPG_ZERO, TH_ATPG_ONE, TH_ATPG_D, TH_ATPG_DBAR, TH_ATPG_X
} thatpg_simv_t;

/* a value assignment to be made on node n */
typedef struct thatpg_assignment {
  node_t *n;
  thatpg_simv_t v;
} thatpg_assignment_t;

/* a faulty node */
typedef struct thfault {
  node_t *node;                     /* node on which fault happens */
  node_t *fanin;                    /* input that is faulty, NIL if output/stem */
  int findex;                       /* fanin index (-1 if primary input/stem fault) */

  thfault_type_t type;              /* type of fault */
  thfault_value_t value;            /* type of SA fault */
  thfault_status_t status;          /* initally set to TH_UNTESTED */

  lsList dfront;                    /* d-frontier for fault */
  int backtracks;                   /* number of backtracks done */
  /* add other info as necessary */
} thfault_t;

/* options data structure of the threshold atpg and faultsim */
typedef struct thatpg_options {
  int verb;                       /* verbosity level */
  char *ifile;                    /* where to read input from */
  char *ofile;                    /* where to dump output to */
  FILE *fpi;                      /* file pointer to read ifile */
  FILE *fpo;                      /* file pointer to write ofile */
  bool help;                      /* print help message */
  
  bool faultdrop;                 /* perform fault dropping during fault simulation */
  bool faultcollapse;             /* perform fault collapsing during fault generation */
  double fcoverage_limit;         /* fault coverage limit when doing random test generation */
  int backtrack_limit;            /* maximum number of backtracks for a fault */
  long int rand_timeout;          /* timeout for random test generation */
  int nvec;                       /* number of vectors to generate each time for fault simulation */
  double don;                     /* delta on parameter */
  double doff;                    /* delta off paramenter */
  
  /* additional user options go here */
} thatpg_options_t;


/* data structure for atpg for threshold networks*/
typedef struct {
  thatpg_options_t *thatpg_opts;  /* all the user options */
  bool seq;                       /* if sequential network */

  st_table *tests;                /* test vectors */
  st_table *fault_table;          /* hash table of faults to test for */

  array_t *simv;                  /* array of BIGINTs containg simulation vectors */
  int nv;                         /* number of simulation vectors */
  array_t *simr;                  /* array of BIGINTs containing fault-free simulation results by executing the vectors in simv */
  array_t *simf;                  /* array of BIGINTs containing the sim results of a particular fault in fault_table */

  int npi;                        /* number of primary inputs */
  int npo;                        /* number of primary outputs */
  int nlatch;                     /* number of latches */
  int nrpi;                       /* number of real primary inputs */
  int nrpo;                       /* number of real primary outputs */
  network_t *network;             /* pointer to the network */

  char *ilpmodel1_file;           /* ilpmodel1 filename */
  char *ilpmodel2_file;           /* ilpmodel2 filename */
  char *ilpdata_file;             /* ilpdata filename */
  char *ilpresult_file;           /* ilpresult filename */

  /* add other information as necessary */
} thatpg_info_t;


/* thcom.c */
EXTERN int is_threshold_network ARGS((network_t *));

/* thconv.c */
EXTERN int th_mapoto ARGS((network_t *, int, int, thsyn_options_t *));

/* thnode.c */
EXTERN thnode_t *thnode_alloc ARGS(());
EXTERN thnode_t *thnode_dup ARGS((thnode_t *));
EXTERN void thnode_free ARGS((thnode_t *));
EXTERN void thnode_print ARGS((FILE *, node_t *, int));
EXTERN void thnode_assign_weights ARGS((node_t *, thsyn_options_t *));
EXTERN double thnode_calc_area ARGS((node_t *, double (*f)()));
EXTERN int thnode_check ARGS((node_t *));
EXTERN void thnode_sim ARGS((node_t *));
EXTERN void thnode_sim32 ARGS((node_t *));
EXTERN void thnode_create_fanin_lines ARGS((node_t *));
EXTERN pset_family *thnode_scovers ARGS((node_t *, st_table *));
EXTERN pset_family *thnode_pcubes ARGS((node_t *, st_table *));




#endif /* TH_H */
