#include <stdio.h>
#include "sis.h"
#include "../speed/speed_int.h"
#include "th_int.h"

static st_table *level_table; /* For the levelize routines */

static int thbuffer_level_cmp ARGS((char **p1, char **p2));
static int thbuffer_print_level ARGS((network_t *network, array_t *node_array, 
				double thresh, int crit_only, int flag));
static int thbuffer_assign_level ARGS((node_t *node));
static void thbuffer_insert_buffers ARGS((network_t *network, node_t *source, node_t *sink, int num_buffer));
static int thbuffer_usage ARGS(());

/***************************************************************************
 * this function does technology mapping for RTD and Single-Electron Boxes *
 * based ciruits by inserting buffers to the circuit wherever needed.      *
 * The circuit's operation is based on the four-phase or three-phase clock *
 * scheme for RTDs and Single-electron boxes respectively.                 *       
 ***************************************************************************/
int com_thbuffer(network_t **network, int argc, char **argv)
{
  int c;
  int level;


  /* parse options */
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "h")) != EOF) {
    switch (c) {
    case 'h': return thbuffer_usage();
    default: break;
    }
  }


  level = thbuffer_print_level(*network, NIL(array_t), 0.5, FALSE, TRUE);

  return TH_SUCCESS;
}


static int thbuffer_print_level(network_t *network, array_t *node_array, 
				double thresh, int crit_only, int flag)
{
    lsGen gen;
    node_t *p;
    char *dummy;
    st_table *node_table;
    delay_time_t slack;
    int i, j, level, new_level;
    array_t *nodevec, *temp_array; 
    int source_level, sink_level, num_buffer, total_buffer = 0;
    node_t *sink_node;

    level_table = st_init_table(st_ptrcmp, st_ptrhash);

    foreach_node(network, gen, p) {
        (void) st_insert(level_table, (char *)p, (char *)0);
    }

    foreach_primary_output(network, gen, p) {
        (void) thbuffer_assign_level(p);
    }

    nodevec = network_dfs_from_input(network);
    array_sort(nodevec, thbuffer_level_cmp);

    if (node_array == NIL(array_t)) {
	/* All nodes can be printed */
	node_table = level_table;
    } else {
	/* Get the nodes to be printed */
	node_table = st_init_table(st_ptrcmp, st_ptrhash);
	for (i = 0; i < array_n(node_array); i++){
	    p = array_fetch(node_t *, node_array, i);
	    (void)st_insert(node_table, (char *)p, NIL(char));
	    temp_array = network_tfi(p, INFINITY);
	    for (j = array_n(temp_array); j-- > 0; ){
		p = array_fetch(node_t *, temp_array, j);
		(void)st_insert(node_table, (char *)p, NIL(char));
	    }
	    array_free(temp_array);
	}
    }

    level = 0;
    if (flag) {
	;
    }
    for (i = 0; i < array_n(nodevec); i++) {
        p = array_fetch(node_t *, nodevec, i);
        if (p->type != PRIMARY_OUTPUT && st_is_member(node_table, (char *)p)) {
	    slack = delay_slack_time(p);
	    if (crit_only && 
		slack.rise > thresh && slack.fall > thresh) continue;
	    if (st_lookup(level_table, (char *)p, &dummy) ){
		new_level = (int)dummy;
		if (new_level > level) {
		    level = new_level;
		    if (flag) {
		      ;
		    }
		}
		if (flag) {
		  ;
		}
		source_level = new_level;
		foreach_fanout(p, gen, sink_node) {
		  if (st_lookup(level_table, (char *)sink_node, &dummy)) {
		    sink_level = (int)dummy;
		    num_buffer = sink_level - source_level - 1;
		    //#ifdef THRESHOLD_SYN_DEBUG0
		    total_buffer = total_buffer + num_buffer;
		    //#endif
		    if (num_buffer < 0)
		      fail("thbuffer.c: number of buffers is negative");
		    /* no need to insert buffer if sink_level - source_level  == 1 */
		    if (num_buffer > 0)
		      thbuffer_insert_buffers(network, p, sink_node, num_buffer);
		  }
		  else 
		    if (node_function(sink_node) != NODE_BUF)
		      fail("thbuffer.c: Node not assigned a level");
		}
	    } else {
		if (flag) 
		  fail("thbuffer.c: Node not assigned a level");
	    }
        }
    }
    if (flag) {
      ;
    }
    array_free(nodevec);
    st_free_table(level_table);
    if (node_array != NIL(array_t)) st_free_table(node_table);

    //#ifdef THRESHOLD_SYN_DEBUG0
    fprintf(sisout, "Network: %s \nTotal_buffers: %d\n", network_name(network), total_buffer);
    //#endif
    return level;
}

static int thbuffer_level_cmp(char **p1, char **p2)
{
    int level1, level2;
    char *dummy;

    (void) st_lookup(level_table, *p1, &dummy);
    level1 = (int) dummy;
    (void) st_lookup(level_table, *p2, &dummy);
    level2 = (int) dummy;
    return (level1 - level2);
}

static int thbuffer_assign_level(node_t *node)
{
    node_t *fanin;
    int i, level, new_level;
    char *dummy;

    (void) st_lookup(level_table, (char *)node, &dummy);
    new_level = (int)dummy;
    if (node->type == PRIMARY_INPUT || new_level > 0) {
        return new_level ;
    } else {
        level = 0;
        foreach_fanin(node, i, fanin) {
            level = MAX(level, thbuffer_assign_level(fanin));
        }
    }
    level++;
    (void) st_insert(level_table, (char *)node, (char *)level);

    return level;
}

/***********************************************************************
 * given the source and sink nodes in a network, this function inverts *
 * 'num_buffer' buffers between these two nodes                        *
 ***********************************************************************/
static void thbuffer_insert_buffers(network_t *network, node_t *source, node_t *sink, int num_buffer)
{
  int i;
  node_t *temp;
  thnode_t *th;
  double *weights;

  if (node_type(sink) == PRIMARY_OUTPUT)
      return;
  for (i = 1; i <= num_buffer; i++) {
    if (i == 1)
      temp = node_literal(source, 1);
    else
      temp = node_literal(temp, 1);
#ifdef THRESHOLD_SYN_DEBUG0
    fprintf(sisout, "\nAdding buffer:\n");
    node_print(sisout, temp);
#endif
    /* if the original network is a threshold network, the 
       new buffer node should also contain threshold and weight 
       information. For now, we just assume the defect tolerance:
       delta_on = 0 and delta_off =1. More general case will be 
       considered later */
    if (network->threshold == 1) {
      /* create thnode */
      th = thnode_alloc();
      th->nin = 1;
      weights = ALLOC(double, 1);
      weights[0] = 1; 
      th->W = weights; /* assuming defect tolerance: delta_on = 0, delta_off = 1 */
      th->T = 1; 
      /* attach the thnode to the node */
      temp->th = (char *) th;
    }
    network_add_node(network, temp);
  }
  node_patch_fanin(sink, source, temp);
}


/******************
 * thbuffer usage *
 ******************/
static int thbuffer_usage(void)
{
  fprintf(siserr, "usage: thbuffer [-h]\n"); 
  fprintf(siserr, "   -h:        \tFor help information. \n");  
  fprintf(siserr, "   This command does technology mapping for RTD and Single-Electron Boxes based ciruits by inserting buffers to the circuit wherever needed. The circuit's operation is based on the four-phase or three-phase clock scheme for RTDs and Single-electron boxes respectively.\n");  
  return 1;
}
