/*=========================================================================*
 *	File    :	thsim.c
 *
 *	Author  :	Pallav Gupta
 *	Created :	Fri Sep 19 20:09:17 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thsim.c,v 1.7 2004/06/21 18:49:09 sis Exp $ */

#include "sis.h"
#include "th_int.h"

static int thsim_usage ARGS(());

/* simulate a threshold network */
int com_thsim(network_t **network, int argc, char **argv)
{
  int i, nvec = 0;
  thsim_options_t *opt;
  FILE *fp;
  array_t *invec, *outvec;

  /* check if threshold network */
  if (is_threshold_network(*network) != TH_SUCCESS)
    return TH_FAIL;

  /* options parse and checks */
  if ((opt = thsim_options_init(argc, argv)) == NIL(thsim_options_t)
      || opt->help) {
    thsim_options_free(opt);
    return thsim_usage();
  }
  
  if (opt->do_stg) {
    fprintf(siserr, "STG simulation not yet implemented. TODO\n");
    thsim_options_free(opt);
    return TH_FAIL;
  }

  if (opt->ip == NIL(FILE)) {
    fprintf(siserr, "Must specify input file to read vectors from\n");
    thsim_options_free(opt);
    return TH_FAIL;
  }

  if (!opt->do_logic && !opt->do_stg) {
    fprintf(siserr, "User suppressed combinational/STG simultion\n");
    thsim_options_free(opt);
    return TH_SUCCESS;
  }

  /* do combinational simulation */
  if (opt->do_logic) {
    /* purely combination network, no latches, we can do parallel simulation */
    if (network_num_latch(*network) == 0) {
      /* read in all test vectors */
      TH_ASSERT((invec = th_read_test_vectors(*network, opt->ip, &nvec)) != NIL(array_t));
      
      /* simulate all the vectors and store result in array */
      outvec = array_alloc(BIGINT **, 0);
      for (i = 0; i < array_n(invec); i++)
	array_insert_last(BIGINT **, outvec,
			  thsim_comb_thnetwork(*network, array_fetch(BIGINT **, invec, i), MAXENTRIES));
	
      fp = (opt->op == NULL) ? sisout : opt->op;      
      fprintf(fp, "Network simulation:\n");
      //fprintf(fp, "Outputs: ");

      /* dump the results */
      for (i = 0; i < array_n(outvec); i++) {
	//fprintf(fp, "Outputs: ");
	if (nvec >= NUMBITS) {
	
	  print_simresults(*network, fp, array_fetch(BIGINT **, outvec, i), NUMBITS);
	}
	else {
	  //	  fprintf(fp, "Outputs: ");
	  print_simresults(*network, fp, array_fetch(BIGINT **, outvec, i), nvec);
	}
	nvec -= NUMBITS;
      }

      /* cleanup space */
      array_bigint_free(invec, network_num_pi(*network));
      array_bigint_free(outvec, network_num_po(*network));
    }
    else { /*TODO how to do parallel FSM simulation with latches?*/
      fprintf(siserr, "Latches contained, Haven't implmeneted simulation\n");
      thsim_options_free(opt);
      return TH_FAIL;
    }
  }

  /* TODO STG SIMULATION */
  if (opt->do_stg) {

  }

  /* free options */
  thsim_options_free(opt);
  return TH_SUCCESS;
}

/* thsim_usage */
static int thsim_usage(void)
{
  fprintf(siserr, "usage: thsim [-c] [-s] [-o <file>] -i <file>\n");
  fprintf(siserr, "    -c \t\tDo not do combinational simulation\n");
  fprintf(siserr, "    -s \t\tDo not do STG simulation (default is NO) (TODO)\n");
  fprintf(siserr, "    -o <file> \tDump results to file\n");
  fprintf(siserr, "    -i <file> \tRead vectors from file\n");
  return TH_FAIL;
}

/* simulate a threshold combinational network and return the result as an array of BIGINT*/
BIGINT **thsim_comb_thnetwork(network_t *network, BIGINT **a, int words)
{
  int i, j, k, l, m, nin, nout;
  array_t *nvec;
  node_t *n;
  void (*sim)();
  lsGen gen;
  BIGINT **out;

  nin = network_num_pi(network);
  nout = network_num_po(network);
  nvec = network_dfs(network);
  sim = thnode_sim32;

  out = ALLOC(BIGINT *, nout);
  for (i = 0; i < nout; i++)
    out[i] = bigint_new();

  /* HACK: this is a terrible hack to attempt to speed up simulation. we only simulate
     nodes whose inputs dont' change. at first i assign all the nodes this magic value.
     reason why i don't use -1 (0xfffffff) or something is because these values are common
     in simulation and if u read the code below u will soon realize that problems might occur.
     but what are the chances u'll get a circuit with a node set to MAGIC value? very low :)
     every now nad then one circuit might fail. but this is not an industrital tool :) */
  foreach_node(network, gen, n) {
    TH_SETVALUE(n, TH_MAGICVALUE);
    TH_SETSIMCHANGED(n, 1);
  }
  
  for (l = 0; l < words; l++) {
    /* set the input values */
    for (i = 0; i < nin; i++)
      TH_SETVALUE(network_get_pi(network, i), bigint_get_word(a[i], l));
    
    /* simulate all internal nodes */
    for (j = 0; j < array_n(nvec); j++) {
      n = array_fetch(node_t *, nvec, j);
      
      /* actually simulate only those nodes whose inputs changed */
      for (m = 0; m < node_num_fanin(n); m++) {
	if (TH_GETSIMCHANGED(node_get_fanin(n, m))) {
	  sim(n);
	  break;
	}
      }
    }
    /* pack the results in bigint */
    for (k = 0; k < nout; k++)
      bigint_set_word(out[k], l, TH_GETVALUE(network_get_po(network, k)));
  }
  
  array_free(nvec);
  return out;
}


/* initialize simulation options data structrure */
thsim_options_t *thsim_options_init(int argc, char **argv)
{
  thsim_options_t *opt;
  int c;
  bool error = FALSE;
  
  opt = ALLOC(thsim_options_t, 1);
  /* default values */
  opt->do_logic = TRUE;
  opt->do_stg = FALSE; /* until we implement FSM simulation TODO */
  opt->help = FALSE;
  opt->ifile = NIL(char);
  opt->ofile = NIL(char);
  opt->ip = NIL(FILE);
  opt->op = NIL(FILE);
  
  /* options parser */
  util_getopt_reset();
  while ((c = util_getopt(argc, argv, "csi:o:h")) != EOF) {
    switch (c) {
    case 'c':
      opt->do_logic = FALSE; break;
    case 's':
      opt->do_stg = FALSE; break;
    case 'i':
      if ((opt->ip = com_open_file(util_optarg, "r", &(opt->ifile), 0)) == NIL(FILE))
	error = TRUE;
      break;
    case 'o': 
      if ((opt->op = com_open_file(util_optarg, "w", &(opt->ofile), 0)) == NIL(FILE))
	error = TRUE;
      break;
    case 'h':
      opt->help = TRUE; break;
    default:
      opt->help = TRUE; break;
    }
  }

  if (error) {
    thsim_options_free(opt);
    return NIL(thsim_options_t);
  }
  else 
    return opt;
}

/* free options data structure */
void thsim_options_free(thsim_options_t *opt)
{
  if (opt == NIL(thsim_options_t))
    return;

  if (opt->ifile != NIL(char))
    FREE(opt->ifile);
  if (opt->ofile != NIL(char))
    FREE(opt->ofile);
  if (opt->op != NIL(FILE)) 
    fclose(opt->op);
  if (opt->ip != NIL(FILE))
    fclose(opt->ip);
  FREE(opt);
}


