/*=========================================================================*
 *	File    :	thutil.c
 *
 *	Author  :	Pallav Gupta
 *	Created :	Tue Sep 30 14:54:11 2003
 *
 * This source code can be redistributed under the GNU Free Public Licence.
 *=========================================================================*/

/* $Id: thutil.c,v 1.8 2004/06/21 18:49:09 sis Exp $ */

#include "sis.h"
#include "th_int.h"

// NOTE: THIS FILE SHOULD CONTAIN THOSE FUNCTIONS THAT AREN'T PARTICULAR TO ANY FILE
// REUSABLE STUFF


/* read test vectors from file and return an array of array of pointers to BIGINTS,
   total number of vectors read is reaturned in n */
array_t *th_read_test_vectors(network_t *network, FILE *fp, int *n)
{
  char *line, *buf;
  int linelen, bitlength, count, nvec;
  array_t *vec;
  BIGINT **a;

  line = NIL(char);
  linelen = nvec = count = 0;
  bitlength = network_num_pi(network) - network_num_latch(network);
  buf = ALLOC(char, NUMBITS * bitlength);
  vec = array_alloc(BIGINT **, 0);
 
  /* read a line */
  while (read_line(fp, &line, &linelen) != EOF) {
    nvec++;
    if (strlen(line) != bitlength) {
      fprintf(siserr, "Wrong number of input values supplied\n");
      FREE(line);
      FREE(vec);
      return NIL(array_t);
    }
    memcpy(&buf[count*bitlength], line, bitlength);
    count++;
    
    /* we have NUMBITS, max size of a BIGINT, add to array */
    if (count == NUMBITS) {
      TH_ASSERT((a = ascii_to_parallel_bigint(buf, bitlength, count, BINARY)) != NIL(BIGINT *));
      array_insert_last(BIGINT **, vec, a);
      count = 0;
    }
  }

  if (count%NUMBITS != 0) {
    TH_ASSERT((a = ascii_to_parallel_bigint(buf, bitlength, count, BINARY)) != NIL(BIGINT *));
    array_insert_last(BIGINT **, vec, a);
  }
 
  TH_ASSERT(n != NIL(int));
  *n = nvec;
  FREE(line);
  FREE(buf);
  return vec;
}


/* return TH_SUCCESS if a threshold network, TH_FAIL otherwsie */
int is_threshold_network(network_t *network)
{
  if (network != NIL(network_t) && network->threshold == 1)
    return TH_SUCCESS;
  else {
    fprintf(siserr, "Not a valid threshold network.\n");
    return TH_FAIL;
  }
}


/* print the result of the simulation */
void print_simresults(network_t *network, FILE *fp, BIGINT **a, int len)
{
  int i, j, max, npo;

  max = (len == NUMBITS) ? NUMBITS : len;
  npo = network_num_po(network);
  for (j = 0; j < max; j++) {
    fprintf(fp, "Outputs: ");
    for (i = 0; i < npo; i++) {
      fprintf(fp, "%d ", bigint_get_bit(a[i], j)); 
    }
    fprintf(fp, "\n");
  }

  /* TODO: FSM */
}


/* compare two arrays of BIGINTS**, return 0 if equal, 1 otherwise */
int compare_sim_results(array_t *a, array_t *b, int n, int nv)
{
  BIGINT **c1, **c2;
  int i, j, status, bits, k; 

  TH_ASSERT(array_n(a) == array_n(b));
  
  for (i = 0, k = nv; i < array_n(a); i++, k -= NUMBITS) {
    c1 = array_fetch(BIGINT **, a, i);
    c2 = array_fetch(BIGINT **, b, i);

    bits = (k >= NUMBITS) ? MAXBITINDEX : (k - 1);
    
    for (j = 0; j < n; j++) {
      TH_ASSERT((status = bigint_compare(c1[j], c2[j], bits)) != BIGINT_ERROR);
      if (status == BIGINT_DIFF)
	return 1;
    }
  }
  
  return 0;
}


/* free an array of BIGINT ** pointers, n is the number of inices in the BIGINT ** */
void array_bigint_free(array_t *a, int n) 
{
  int i;

  if (a == NIL(array_t))
    return;
  
  for (i = 0; i < array_n(a); i++)
    bigint_array_free(array_fetch(BIGINT **, a, i), n);
  array_free(a);
}


/* generate a vector of length bits in binary format*/
char *random_bitvector(int length)
{
  char *s, *p, *k;
  int i, num;
  int size = sizeof(int);

  TH_ASSERT(length > 0);

  num = length/(2 * size * size);
  if (length % (2 * size * size))
    num++;

  s = ALLOC(char, 2 * num * size);
  for (i = 0, p = s; i < num; i++, p += (2 * size))
   sprintf(p, "%08x", rand());

  k = ALLOC(char, BYTESIZE * num * size);
  for (i = (2 * num * size) - 1, p = k; i >= 0; i--, p += 4) {
    switch (s[i]) {
    case 'a': memcpy(p, "1010", 4); break;
    case 'b': memcpy(p, "1011", 4); break;
    case 'c': memcpy(p, "1100", 4); break;
    case 'd': memcpy(p, "1101", 4); break;
    case 'e': memcpy(p, "1110", 4); break;
    case 'f': memcpy(p, "1111", 4); break;
    case '0': memcpy(p, "0000", 4); break;
    case '1': memcpy(p, "0001", 4); break;
    case '2': memcpy(p, "0010", 4); break;
    case '3': memcpy(p, "0011", 4); break;
    case '4': memcpy(p, "0100", 4); break;
    case '5': memcpy(p, "0101", 4); break;
    case '6': memcpy(p, "0110", 4); break;
    case '7': memcpy(p, "0111", 4); break;
    case '8': memcpy(p, "1000", 4); break;
    case '9': memcpy(p, "1001", 4); break;
    default:
      fail("bad hex character");
    }
  }
  
  p = ALLOC(char, length+1);
  memset(p, 0, length+1);
  memcpy(p, k, length);

  FREE(s);
  FREE(k);

  return p;
}


/* generate a seed based on time (got it off the net somewhere) */
unsigned time_seed(void) 
{
  time_t timeval;
  unsigned char *ptr;
  unsigned seed;
  size_t i;

  timeval = time (NULL);
  ptr = (unsigned char *) &timeval;

  seed = 0;
  for (i = 0; i < sizeof(timeval); i++)
    seed = seed * (UCHAR_MAX + 2U) + ptr[i];

  return seed;
}
